import unittest
from mystix.language.shared.ast.Var import Var
from mystix.language.shared.ast.Value import Value
from mystix.language.shared.ast.math_ast.Operand import Operand
from mystix.language.shared.ast.math_ast.MathFuncs import MathFuncs
from mystix.language.shared.ast.math_ast.SimpleFunc import SimpleFunc
from mystix.language.shared.primitives import numerical, values
from mystix.language.evaluation.evaluator import Evaluator

class TestMathFuncs(unittest.TestCase):

    pass


import unittest
from typing import List


from mystix.language.shared import ast
from mystix.language.parsing import Parser
from mystix.language.tokenization import Tokenizer, TokenizationError
from mystix.language.evaluation import Evaluator
from mystix.language.shared.primitives import Types, values, graphs
from mystix.ui.util import read_program_file
from tests.util import ast_equal


class CompilePipelineTests(unittest.TestCase):
    """
    Test tokenization through evaluation
    useful right now for validating function stubs
    """

    def test(self):
        content = "NOT A REAL PROGRAM"
        try:
            t = Tokenizer(content)
            t.tokenize()
            program: ast.Program = Parser(t).parseProgram()
            result: int = Evaluator(graphics=False).evaluate(program, duration=5000)
            self.fail()
        except TokenizationError:
            pass

    def test_regular(self):
        p_expected: ast.Program = ast.Program(
            ast.Body([

                # source = live remote "https://covid-api.com/api/reports"
                ast.Loader(ast.Var("source"),
                           ast.Source("https://covid-api.com/api/reports")),

                # map source "confirmed" to number confirmed
                ast.Mapper(ast.Var("source"), "confirmed",
                           ast.Declare(ast.Type(Types.NUMBER), ast.Var("confirmed"))),

                # number count = 0
                ast.Assigner(ast.Declare(ast.Type(Types.NUMBER), ast.Var("count")),
                             ast.Value(values.IntegerValue(0))),

                # observe(source) do count++
                ast.Trigger(ast.Var("source"), ast.MathFuncs([ast.Increment(ast.Var(
                    "count"))])),

                # plot line_xy(count,confirmed) called "confirmed_cases"
                ast.Plotter(ast.Graph(graphs.LineXYGraph()),
                            ast.VarAxis(ast.Var("count")),
                            ast.VarAxis(ast.Var("confirmed")), "confirmed_cases"),

            ]))

        def parse(content) -> ast.Program:
            t = Tokenizer(content)
            t.tokenize()
            return Parser(t).parseProgram()

        p: ast.Program = read_program_file("tests/res/programs/regular_program.mstx",
                                           parse)

        self.assertTrue(ast_equal(p_expected, p))



from typing import List
import re


class TokenizationError(Exception):
    pass


class Tokenizer:

    def __init__(self, content: str):
        self.current_token = 0
        self.content = content
        self.fixed_literals = [
            ',', ';', '+=', '-=', '/=', '^=', '*=', 'log', 'sin', 'cos', 'exp',
            'line_xy', 'scatter_xy', 'true', 'false', 'number', 'category',
            'binary', 'remote', 'plot', 'titled', '(', ')', 'do',
            'observe', 'map', 'to', 'program:', 'start!', '=', ' ', '"'
        ]
        self.tokens: List[str] = []
        self.tokenize()

    def tokenize(self) -> List[str]:

        RESERVED_WORD = "$%$"

        tokenized_program = self.content.replace("\n", ";")
        for p in self.fixed_literals:
            tokenized_program = tokenized_program.replace(
                p, RESERVED_WORD + p + RESERVED_WORD)
        tokenized_program = tokenized_program.replace(
            RESERVED_WORD+RESERVED_WORD, RESERVED_WORD)
        if len(tokenized_program) > 0 and tokenized_program.startswith(RESERVED_WORD):
            tokenized_program = tokenized_program[len(RESERVED_WORD):]
        res = tokenized_program.split(RESERVED_WORD)

        # Stretch goal: Optimize with less iterative space/time if possible solution exists
        new_res = []
        for i in res:
            if i.strip():
                new_res.append(i.strip())
        self.tokens = new_res
        return new_res

    def check_next(self):
        if self.current_token < len(self.tokens):
            return self.tokens[self.current_token]
        else:
            return "NO MORE TOKENS"

    def get_next(self):
        if self.current_token < len(self.tokens):
            token = self.tokens[self.current_token]
            self.current_token += 1
        else:
            token = "NULLTOKEN"
        return token

    def check_token(self, regexp) -> bool:
        s = self.check_next()
        if re.match(regexp, s) is not None:
            return True
        return False

    def get_and_check_next(self, regexp):
        s = self.get_next()
        if (re.match(regexp, s) == None):
            raise TokenizationError(
                f"Unexpected token! Expected something matching {regexp}, got {s}")
        return s

    def more_tokens(self) -> bool:
        if len(self.tokens) == 0:
            return False
        return self.current_token < len(self.tokens)
        
    def get_line(self, stopper = ';'):
        line = []
        cur_token = self.current_token
        next_token = self.get_next()
        stoppers = [';', stopper]
        while(next_token not in stoppers):
            line.append(next_token)
            next_token = self.get_next()
        self.current_token = cur_token
        return line



from .Command import Command
from ..graphs_ast.Graph import Graph
from ..graphs_ast.Axis import Axis
from mystix.language.shared.ast.visitor import Visitor


class Plotter(Command):

    def __init__(self, graph: Graph, x_axis: Axis, y_axis: Axis, name: str):
        self.graph: Graph = graph
        self.x: Axis = x_axis
        self.y: Axis = y_axis
        self.graph_name: str = name

    def accept(self, v: Visitor):
        return v.visit_plotter(self)



from .Node import Node
from .Type import Type
from .Var import Var
from mystix.language.shared.ast.visitor import Visitor


class Declare(Node):

    def __init__(self, t: Type, v: Var):
        self.type = t
        self.var = v

    def accept(self, v: Visitor):
        return v.visit_declare(self)




import unittest as ut
from typing import List
from mystix.ui.util import read_program_file
from mystix.language.tokenization import Tokenizer, TokenizationError
from mystix.language.shared.ast import Program


class TestTokenizer(ut.TestCase):

    def test_Tokenizer_1(self):
        def run_compile(content: str) -> List[str]:
            return Tokenizer(content).tokenize()
        p: List[str] = read_program_file(
            "tests/res/programs/example1", run_compile)
        res = ['program:', ';', 'source', '=', 'remote', "(",
               '"', 'www.coviddata.com/stream', '"', ")",';', 'map', "(",
               'source', ")", '"', 'case_date', '"', 'to', 'number', 'date', ';', 'number',
               'count', '=', '0', ';',
               'observe', "(", 'source', ")", "do", 'count++', ';', 'plot',
               'scatter_xy',"(" ,
               'date', ",", 'age', ")",
               'titled', '"', 'age_graph', '"', ';',
               'plot', 'line_xy', "(",'date', ",",'log','(', 'count', ')',")",
               'titled',
               '"', 'cases_', 'log', '"', ';', 'start!']
        self.assertEqual(res, p)

    def test_Tokenizer_2(self):
        def run_compile(content: str) -> List[str]:
            return Tokenizer(content).tokenize()
        p: List[str] = read_program_file(
            "tests/res/programs/example2", run_compile)
        res = ['<START>', ';', 'source', '=',
               '"', 'http://winterolympicsmedals.com/medals.csv', '"',
               ';', 'map', 'source', '"', 'Year', '"', 'to', 'number', 'year', ';',
               'map', 'source', '"', 'Medal', '"', 'to', 'number',
               'medal', ';',
               'year', '=', 'year','/', '100', ';',
               'year', '=', 'year','*', '10', ';',
               'year', '=', 'year','+', '23', ';',
               'year', '=', 'year','-', '12', ';',
               'plot', 'xy', 'year', 'medal', 'titled', 'medal_graph', ';', '<END>']
        self.assertEqual(res, p)

    def test_Tokenizer_3(self):
        def run_compile(content: str) -> List[str]:
            return Tokenizer(content).tokenize()
        p: List[str] = read_program_file(
            "tests/res/programs/example3", run_compile)
        res = ['<START>', ';', 'source', '=', 'static', 'remote',
               '"', 'http://winterolympicsmedals.com/medals.csv','"',
               ';', 'map', 'source', '"', 'Year', '"', 'to', 'number', 'year', ';',
               'map', 'source', '"', 'Medal', '"', 'to', 'number',
               'medal', ';', 'number', 'shortyear', '=', 'year', '-', '1900', ';', 'number', 'addyear', '=', 'year',
               '+', '50', ';', 'number', 'timesyear', '=', 'year', '*', '10', ';', 'number', 'divyear', '=', 'year',
               '/', '100', ';', 'number', 'sin',
               'year',
               '=',
               'sin',
               '(',
               'year',
               ')',
               ';',
               'number',
               'cos',
               'year',
               '=',
               'cos',
               '(',
               'year',
               ')',
               ';',
               'number',
               'exp',
               'year',
               '=',
               'exp',
               '(',
               'year',
               ')', ';', 'number', 'sqyear', '=', 'year', '^', '2', ';', 'number',
               'thing', '=', '2000', ';', 'thing--', ',', 'addyear++', ';', 'binary',
               'bin', '=', 'true', ';', 'binary', 'bin2', '=',
               'false', ';', 'plot', 'xy', 'cos', '(', 'year', ')', 'medal', 'titled',
                                    'cos', '_medal_graph', ';', '<END>']
        self.assertEqual(res, p)

    def test_Tokenizer_get_next(self):
        def run_compile(content: str) -> str:
            return Tokenizer(content).get_next()
        def run_compile2(content: str) -> str:
            token_obj = Tokenizer(content)
            token_obj.get_next()
            return token_obj.get_next()
        def run_compile3(content: str) -> str:
            token_obj = Tokenizer(content)
            token_obj.get_next()
            token_obj.get_next()
            return token_obj.get_next()
        p: str = read_program_file("tests/res/programs/example1", run_compile)
        self.assertEqual('program:', p)
        p = read_program_file("tests/res/programs/example1", run_compile2)
        self.assertEqual(';', p)
        p = read_program_file("tests/res/programs/example1", run_compile3)
        self.assertEqual('source', p)

    def test_Tokenizer_check_next(self):
        def run_compile(content: str) -> str:
            token_obj = Tokenizer(content)
            return token_obj.check_next()
        p: str = read_program_file("tests/res/programs/example1", run_compile)
        self.assertEqual('program:', p)
        
    def test_Tokenizer_check_next2(self):
        def run_compile(content: str) -> str:
            return Tokenizer(content).check_next()
        p: str = read_program_file("tests/res/programs/empty", run_compile)
        self.assertEqual("NO MORE TOKENS", p)

    def test_Tokenizer_check_token(self):
        def run_compile(content: str) -> bool:
            token_obj = Tokenizer(content)
            return token_obj.check_token('program:')
        def run_compile2(content: str) -> bool:
            token_obj = Tokenizer(content)
            token_obj.get_next()
            token_obj.get_next()
            return token_obj.check_token('program:')
        p: bool = read_program_file("tests/res/programs/example1", run_compile)
        self.assertTrue(p)
        p = read_program_file("tests/res/programs/example1", run_compile2)
        self.assertFalse(p)

    def test_Tokenizer_get_and_check_next(self):
        def run_compile(content: str) -> str:
            return Tokenizer(content).get_and_check_next('program:')
        p: str = read_program_file("tests/res/programs/example1", run_compile)
        self.assertEqual('program:', p)

    def test_Tokenizer_get_and_check_next2(self):
        def run_compile(content: str):
            try:
                Tokenizer(content).get_and_check_next('source')
                self.fail()
            except TokenizationError:
                pass
        read_program_file("tests/res/programs/example1", run_compile)

    def test_Tokenizer_more_tokens(self):
        def run_compile(content: str) -> bool:
            return Tokenizer(content).more_tokens()
        p: bool = read_program_file("tests/res/programs/example1", run_compile)
        self.assertTrue(p)

    def test_Tokenizer_more_tokens2(self):
        def run_compile(content: str) -> bool:
            return Tokenizer(content).more_tokens()
        p: bool = read_program_file("tests/res/programs/empty", run_compile)
        self.assertFalse(p)



from .Command import Command
from ..Var import Var
from ..Declare import Declare
from mystix.language.shared.ast.visitor import Visitor


class Mapper(Command):

    def __init__(self, src: Var, tbl_field: str, decl: Declare):
        self.src: Var = src
        self.tbl_field: str = tbl_field
        self.decl = decl

    def accept(self, v: Visitor):
        return v.visit_mapper(self)





from mystix.language.shared import ast
from mystix.language.shared.primitives import Types
from mystix.language.shared.primitives.values import IntegerValue
from mystix.language.shared.primitives.graphs import ScatterXYGraph, LineXYGraph
from mystix.language.shared.primitives.numerical import NumFunction


def example_1() -> ast.Program:
    """
    Instantiates program example 1 as a native AST
    """
    return ast.Program(
        ast.Body([

            # source = live remote "www.coviddata.com/stream"
            ast.Loader(ast.Var("source"),
                                    ast.Source("www.coviddata.com/stream")),

            # map source "case_date" to number date
            ast.Mapper(ast.Var("source"), "case_date",
                       ast.Declare(ast.Type(Types.NUMBER), ast.Var("date"))),

            # number count = 0
            ast.Assigner(ast.Declare(ast.Type(Types.NUMBER), ast.Var("count")),
                         ast.Value(IntegerValue(0))),

            # on new data from source count++
            ast.Trigger(ast.Var("source"), ast.MathFuncs([ast.Increment(ast.Var(
                "count"))])),

            # plot xy date age titled age_graph
            ast.Plotter(ast.Graph(ScatterXYGraph()),
                        ast.VarAxis(ast.Var("date")),
                        ast.VarAxis(ast.Var("age")), "age_graph"),

            # plot line xy date log(count) titled cases_log
            ast.Plotter(ast.Graph(LineXYGraph()),
                        ast.VarAxis(ast.Var("date")),
                        ast.FuncAxis(ast.BuiltinFunc(NumFunction.LOG, ast.Var(
                            "count"))),
                        "age_graph"),
        ]))


def simple_plot_example() -> ast.Program:

    return ast.Program(ast.Body([
        ast.Plotter(ast.Graph(ScatterXYGraph()),
                    ast.VarAxis(ast.Var("t")),
                    ast.FuncAxis(ast.BuiltinFunc(NumFunction.SIN, ast.Var(
                        "t"))),
                    "sine_wave"),

    ]))



from collections import defaultdict
import random

from mystix.language.shared.primitives.values import Value, IntegerValue
from mystix.language.evaluation.errors import LanguageError


class OutOfMemoryError(LanguageError):
    pass


class SegmentationFault(LanguageError):
    pass


def _rand_mem_() -> Value:
    # simulate uninitialized memory
    return IntegerValue(random.randint(0, 100))


class Memory:

    MAX_VALUES = 2000   # memory limit

    def __init__(self):
        self._map_ = defaultdict(
            lambda: _rand_mem_())

    def get_fresh_loc(self) -> int:
        i = 0
        while i < self.MAX_VALUES:
            if i in self._map_:
                i = i + 1
            else:
                return i
        raise OutOfMemoryError()

    def _valid_loc_(self, loc: int) -> bool:
        return 0 <= loc < self.MAX_VALUES

    def write(self, loc: int, val: Value) -> Value:
        if not self._valid_loc_(loc):
            raise SegmentationFault
        else:
            self._map_[loc] = val
            return val

    def read(self, loc: int) -> Value:
        if not self._valid_loc_(loc):
            raise SegmentationFault
        else:
            return self._map_[loc]

    def free(self, loc: int) -> None:
        if not self._valid_loc_(loc):
            raise SegmentationFault
        else:
            self._map_.pop(loc)


class UndefinedVariableError(LanguageError):
    def __init__(self, var_name: str):
        super(LanguageError, self).__init__(
            "Variable '%s' was referenced out of scope!" % var_name)


class Environment:

    def __init__(self):
        self._mem_ = Memory()
        self._map_ = defaultdict(lambda: [])

    def extend(self, name: str, val: Value) -> int:
        l: int = self._mem_.get_fresh_loc()
        self._mem_.write(l, val)
        self._map_[name].append(l)
        return l

    def get_val(self, name: str) -> Value:
        locations = self._map_[name]
        q = len(locations)
        if q == 0:
            raise UndefinedVariableError(name)
        else:
            return self._mem_.read(locations[q-1])

    def set_val(self, name: str, val: Value) -> Value:
        locations = self._map_[name]
        q = len(locations)
        if q == 0:
            raise UndefinedVariableError(name)
        else:
            return self._mem_.write(locations[q-1], val)

    def undef(self, name) -> None:
        locations = self._map_[name]
        q = len(locations)
        if q == 0:
            raise UndefinedVariableError(name)
        else:
            l: int = locations[q-1]
            self._map_[name] = locations[:q-1]
            self._mem_.free(l)



from typing import cast


class Value:

    def equals(self, v) -> bool:
        return False


class NumericalValue(Value):

    def __init__(self):
        self.value = 0

    def equals(self, v: Value) -> bool:
        if isinstance(v, NumericalValue):
            return self.value == cast(NumericalValue, v).value
        else:
            return False


class CategoricalValue(Value):

    def __init__(self, val: str):
        self.value = val

    def equals(self, v: Value) -> bool:
        if isinstance(v, CategoricalValue):
            return self.value == cast(CategoricalValue, v).value
        else:
            return False


class BinaryValue(Value):

    def __init__(self, val: bool):
        self.value: bool = False

    def equals(self, v: Value) -> bool:
        if isinstance(v, BinaryValue):
            return self.value == cast(BinaryValue, v).value
        else:
            return False


class IntegerValue(NumericalValue):

    def __init__(self, val: int):
        super(NumericalValue, self).__init__()
        self.value = val

    def __repr__(self):
        return "%d" % self.value


class FloatValue(NumericalValue):

    def __init__(self, val: float):
        super(NumericalValue, self).__init__()
        self.value = val



from typing import List

from .Node import Node
from .commands_ast.Command import Command
from mystix.language.shared.ast.visitor import Visitor


class Body(Node):
    """
    Program body
    """
    def __init__(self, cs: List[Command]):
        self.commands: List[Command] = cs
        

    def accept(self, v: Visitor):
        return v.visit_body(self)


import unittest


from tests.util.example_ast import example_1
from tests.util import ast_equal

from mystix.ui.util import read_program_file
from mystix.language.tokenization import Tokenizer
from mystix.language.parsing import Parser
from mystix.language.shared.ast import Program


class ExampleParseTests(unittest.TestCase):

    def test_parse_1(self):
        def run_compile(content: str) -> Program:
            t = Tokenizer(content)
            t.tokenize()
            return Parser(t).parseProgram()
        p: Program = read_program_file("tests/res/programs/example1", run_compile)
        #self.assertTrue(ast_equal(p, example_1()))



from mystix.ui.graphics import Graphics
from typing import Callable, List
from os import path, listdir, remove, mkdir
import pickle
import re
from pyqtgraph import PlotItem  # type: ignore

from mystix.language.evaluation.errors import LanguageError


class GraphManagerError(LanguageError):
    pass


CONST_FRAGMENT_CACHE = "/tmp/mystix_vis_cache"
CONST_DEBUG_OUTPUT = "/tmp/mystix_vis_trace"


class Plot:

    CONST_DATA_LIMIT = 10000

    def __init__(self, name: str, plot: PlotItem, line: bool):
        self.name = name
        self.fr_fmt = re.compile("%s_([0-9]+)" % self.name)
        self.graphics: PlotItem = plot
        self.plot_data \
            = self.graphics.plot() if line \
            else self.graphics.plot(pen=None, symbol='x', symbolPen=None,
                                    symbolSize=10)
        self.data_frag = 0

        self.x_data: List = []
        self.y_data: List = []

    def update(self):
        fragments = sorted([
            int(self.fr_fmt.match(fn).group(1)) for fn in listdir(
                CONST_FRAGMENT_CACHE)
            if path.isfile(path.join(CONST_FRAGMENT_CACHE, fn))
            and self.name in fn])
        for fr in fragments:
            fn = self.num_fragment(fr)
            with open(fn, "rb") as f:
                try:
                    x_data, y_data = pickle.loads(f.read())
                    self.x_data = (self.x_data + x_data)
                    self.y_data = (self.y_data + y_data)
                    # data limiting
                    if len(self.x_data) > Plot.CONST_DATA_LIMIT:
                        self.x_data = (self.x_data + x_data)[-Plot.CONST_DATA_LIMIT:]
                    if len(self.y_data) > Plot.CONST_DATA_LIMIT:
                        self.y_data = (self.y_data + y_data)[-Plot.CONST_DATA_LIMIT:]
                    read = True
                except EOFError:
                    continue
            if read:
                remove(fn)
        self.plot_data.setData(self.x_data, self.y_data)

    def num_fragment(self, i: int):
        return path.join(CONST_FRAGMENT_CACHE, self.name + "_%d" % i)

    def log(self, msg: str):
        with open(path.join(CONST_DEBUG_OUTPUT, self.name + ".log"), "a") as trace:
            trace.write("[]: %s\n" % msg)

    def name_fragment(self):
        return self.num_fragment(self.data_frag)

    def add_data(self, x_data: List, y_data: List):
        if len(x_data) != len(y_data):
            raise GraphManagerError("New data arrays are not the same length")
        with open(self.name_fragment(), "wb") as frag:
            frag.write(pickle.dumps((x_data,y_data)))
        self.data_frag = self.data_frag + 1

    def clear_cache(self):
        self.log("Clearing cache")
        files = [path.join(CONST_FRAGMENT_CACHE, f) for f in listdir(
            CONST_FRAGMENT_CACHE) if
                 self.fr_fmt.match(f)]
        for f in files:
            remove(f)


class GraphManager:

    def __init__(self):
        self.closed = False
        self.graphics = Graphics()
        self.plots = {}
        self.graphics.add_window("Mystix", 600, 600)
        self.graphics.add_update(lambda: self.update_plots())
        for d in [CONST_DEBUG_OUTPUT, CONST_FRAGMENT_CACHE]:
            if path.isfile(d):
                raise GraphManagerError("%s is already a file" % d)
            if not path.isdir(d):
                mkdir(d)

    def add_plot(self, plot_name: str, line_plot: bool = False):
        if ' ' in plot_name:
            raise GraphManagerError("Plot name cannot contain spaces")
        if plot_name in self.plots:
            raise GraphManagerError("Plot '%s' already exists" % plot_name)
        else:
            p: Plot = Plot(plot_name, self.graphics
                           .get_window("Mystix")
                           .addPlot(title=plot_name), line_plot)
            self.plots[plot_name] = p

    def update_plots(self):
        if not self.graphics.to_close and not self.closed:
            try:
                for p in self.plots:
                    self.plots[p].update()
            except RuntimeError:
                print("Warning: plots attempted to update after close")

    def _confirm_plot_(self,  plot_name: str, p: Callable):
        if plot_name not in self.plots:
            raise GraphManagerError("Plot '%s' does not exist" % plot_name)
        else:
            return p(self.plots[plot_name])

    def add_plot_data(self, plot_name: str, x_data: List, y_data: List):
        def add(p: Plot):
            p.add_data(x_data, y_data)
        return self._confirm_plot_(plot_name, add)

    def clean(self):
        for p in self.plots.values():
            p.clear_cache()

    def close(self):
        print("Closing graph manager")
        self.closed = True
        self.graphics.close()



import unittest
from mystix.targets.data import DataLoader, DataLoaderError
import filecmp
from typing import List


class DataLoaderTests(unittest.TestCase):

    # def test_run(self):
    #     load_data("http://winterolympicsmedals.com/medals.csv",
    #               "tests/out/dataLoaderOutputActual.txt")
    #     self.assertTrue(filecmp.cmp('tests/res/dataLoaderOutputExpected.txt',
    #                                 'tests/out/dataLoaderOutputActual.txt'))

    def test_not_exist(self):
        d = DataLoader()
        try:
            d.get_new("data")
            self.fail()
        except DataLoaderError:
            pass

    def test_register_already_exist(self):
        d = DataLoader()
        d.register_source("awkjlda", "data")
        try:
            d.register_source("awkjldadaw", "data")
            self.fail()
        except DataLoaderError:
            pass

    def test_get_new_invalid_host(self):
        d = DataLoader()
        d.register_source("https://covid-apiiii.com/api/reports", "covid_api")
        try:
            d.get_new("covid_api")
            self.fail()
        except DataLoaderError as e:
            pass

    def test_get_new_invalid_request_fmt(self):
        d = DataLoader()
        d.register_source("https://api.covid19api.com/premium/country/south-africa",
                          "covid_api")
        try:
            d.get_new("covid_api")
            self.fail()
        except DataLoaderError as e:
            pass

    def test_get_new_invalid_response_fmt(self):
        d = DataLoader()
        d.register_source("https://google.com", "source")
        try:
            d.get_new("source")
            self.fail()
        except DataLoaderError as e:
            pass

    def test_get_new(self):
        d = DataLoader()
        d.register_source("https://covid-api.com/api/reports", "covid_api")
        data: List = d.get_new("covid_api")
        self.assertTrue(len(data) > 0)  # assumed only for mock
        self.assertTrue('confirmed' in data[0])

    def test_multi_source(self):
        d = DataLoader()
        d.register_source("https://covid-api.com/api/reports", "covid_api")
        d.register_source("https://covid-api.com/api/reports", "covid_api_2")
        data: List = d.get_new("covid_api")
        data2: List = d.get_new("covid_api_2")
        self.assertEqual(data, data2)

    def test_multi_get(self):
        d = DataLoader()
        d.register_source("https://covid-api.com/api/reports", "covid_api")
        data: List = d.get_new("covid_api")
        data2: List = d.get_new("covid_api")
        self.assertNotEqual(data, data2)



import unittest

from mystix.language.evaluation.vars import Memory, OutOfMemoryError
from mystix.language.shared.primitives.values import IntegerValue, Value


class MemoryTests(unittest.TestCase):

    def test_fresh_location_no_set(self):
        m = Memory()
        l1 = m.get_fresh_loc()
        l2 = m.get_fresh_loc()
        self.assertEqual(l1,l2)

    def test_fresh_location_post_set(self):
        m = Memory()
        l1 = m.get_fresh_loc()
        m.write(l1, Value())
        l2 = m.get_fresh_loc()
        self.assertNotEqual(l1, l2)

    def test_set_get(self):
        m = Memory()
        l1 = m.get_fresh_loc()
        m.write(l1, IntegerValue(22))
        v: Value = m.read(l1)
        self.assertTrue(isinstance(v, IntegerValue))
        self.assertEqual(v.value, 22)

    def test_free_location_post_set(self):
        m = Memory()
        l1 = m.get_fresh_loc()
        m.write(l1, Value())
        m.free(l1)
        l2 = m.get_fresh_loc()
        self.assertEqual(l1, l2)

    def test_overwrite(self):
        m = Memory()
        l1 = m.get_fresh_loc()
        m.write(l1, IntegerValue(22))
        m.write(l1, IntegerValue(23))
        v: Value = m.read(l1)
        self.assertTrue(isinstance(v, IntegerValue))
        self.assertEqual(v.value, 23)

    def test_no_limit_on_fresh_no_write(self):
        m = Memory()
        for i in range(Memory.MAX_VALUES):
            m.get_fresh_loc()
        try:
            m.get_fresh_loc()
        except OutOfMemoryError:
            self.fail()

    def test_memory_limit(self):
        m = Memory()
        for i in range(Memory.MAX_VALUES):
            l = m.get_fresh_loc()
            m.write(l, IntegerValue(1))
        try:
            m.get_fresh_loc()
            self.fail()
        except OutOfMemoryError:
            pass



from ..Node import Node
from ..Var import Var
from ..math_ast.MathFuncs import Func
from mystix.language.shared.ast.visitor import Visitor


class Axis(Node):

    def accept(self, v: Visitor):
        pass


class VarAxis(Axis):

    def __init__(self, v: Var):
        self.var = v

    def accept(self, v: Visitor):
        return v.visit_var_axis(self)


class FuncAxis(Axis):

    def __init__(self, f: Func):
        self.fun = f

    def accept(self, v: Visitor):
        return v.visit_func_axis(self)



from __future__ import annotations
from typing import TYPE_CHECKING as STATIC_CHECK, List, Callable, Optional
if STATIC_CHECK:
    from mystix.language.shared.ast import *
from mystix.language.shared.ast.visitor import Visitor
from mystix.language.shared.primitives import values, numerical
from mystix.language.shared.primitives.graphs import (
    Graph as ConcreteGraphType, LineXYGraph)
from mystix.targets.visualization.graphs import GraphManager
from concurrent.futures import ThreadPoolExecutor
import time
from mystix.targets.data import DataLoader
from mystix.language.shared.primitives.types import Types
from mystix.targets.analysis.math import apply_fn, apply_op, apply_qk
from .errors import LanguageError
from .vars import Environment
from collections import defaultdict


TYPES = {
    Types.NUMBER: values.NumericalValue,
    Types.BINARY: values.BinaryValue,
    Types.CATEGORY: values.CategoricalValue
}


class Evaluator(Visitor):

    def __init__(self, graphics: bool = False):
        """
        :param graphics: Controls graphics functionality
        if set to False, any behaviour involving graphics
        will simply not occur.
        """
        self.env = Environment()
        self.gm = GraphManager() if graphics else None
        self.plots: List = []
        self.data = DataLoader()
        # maps a source name to a field mapping dict
        self.maps: defaultdict = defaultdict(lambda: {})
        # for source name provide a list of procedures to execute
        self.events: defaultdict = defaultdict(lambda: [])
        self.sources: List[str] = []
        self.stopped = False  # TODO: use mutex?

    def do_graphics(self, p: Callable):
        if self.gm is not None:
            p(self.gm)

    def update_plots(self):
        # vars -> plots
        for pl in self.plots:
            g: str = pl[0]
            a1: Axis = pl[1]
            a2: Axis = pl[2]
            x = a1.accept(self)
            y = a2.accept(self)

            def add(m: GraphManager):
                m.add_plot_data(g, [x.value], [y.value])
            self.do_graphics(add)

    def update_sources(self):
        # loader -> vars
        for s in self.sources:
            data: List[dict] = self.data.get_new(s)
            maps = self.maps[s]
            for d in data:
                rs = [e() for e in self.events[s]]
                for f in maps:
                    if f in d:
                        val = self.env.get_val(maps[f])
                        d_val = d[f]
                        if isinstance(val, values.NumericalValue):
                            if type(d_val) is int:
                                val = values.IntegerValue(d_val)
                            elif type(d_val) is float:
                                val = values.FloatValue(d_val)
                            else:
                                raise LanguageError(
                                    "Cannot map %s to number %s"
                                    % (str(d_val), maps[f]))
                        elif isinstance(val, values.BinaryValue):
                            if type(d_val) is bool:
                                val = values.BinaryValue(d_val)
                            else:
                                raise LanguageError(
                                    "Cannot map %s to binary %s"
                                    % (str(d_val), maps[f]))
                        else:  # CategoricalValue
                            if type(d_val) is not dict and type(d_val) is not list:
                                val = values.CategoricalValue(str(d_val))
                            else:
                                raise LanguageError("Cannot map %s to %s"
                                                    % (str(d_val), maps[f]))
                        self.env.set_val(maps[f], val)
                        self.update_plots()
                    else:
                        raise LanguageError("Source %s does not have a field %s"
                                            % (s, f))

    def execute(self, p: Program, duration: Optional[int]):
        # internal clock
        t_init = time.time()
        update_interval = 3
        if duration is not None:
            for i in range(int(duration/(1000*update_interval))):
                self.update_sources()
                if (time.time()-t_init) * 1000 > duration or self.stopped:
                    break
                else:
                    time.sleep(update_interval)
        else:
            while not self.stopped:
                self.update_sources()
                time.sleep(update_interval)
        return 0, None

    def evaluate(self, p: Program, duration=None):
        """
        :param p: Program to evaluate
        :param duration: length of time to execute for (in milliseconds).
        A None value will result in program running until graphics are closed.
        """

        # perform the initial traversal
        try:
            self.visit_program(p)
        except LanguageError as e:
            print("\nERROR: %s\n" % str(e))
            return 1, e

        exit_val = 0

        def runtime():
            try:
                return self.execute(p, duration)
            except LanguageError as e:

                def close(m: GraphManager):
                    m.close()
                self.do_graphics(close)

                return 1, e
        with ThreadPoolExecutor() as threads:
            logic = threads.submit(runtime)

            def start(m: GraphManager):
                m.graphics.display(ttl=duration)
            self.do_graphics(start)
            def stop(m: GraphManager):
                self.stopped = True
            self.do_graphics(stop)
            exit_val, err = logic.result()
            print("Program execution completed!")
            if exit_val != 0:
                print("\nERROR: %s\n" % str(err))

            def clean(m: GraphManager):
                m.clean()
            self.do_graphics(clean)

        return exit_val, err

    def visit_program(self, p: Program):
        p.body.accept(self)

    def visit_body(self, b: Body):
        for c in b.commands:
            c.accept(self)
    
    def visit_loader(self, l: Loader):
        # TODO: add source as a value in environment?
        v: Var = l.name
        self.data.register_source(l.source.accept(self), v.name)
        self.sources.append(v.name)

    def visit_mapper(self, m: Mapper):
        v_name = m.decl.accept(self)
        # TODO: source should be stored in environment
        s_name = m.src.name
        self.maps[s_name][m.tbl_field] = v_name

    def visit_declare(self, d: Declare) -> str:
        t: Types = d.type.accept(self)
        # Enumerate the type cases
        c = TYPES[t]
        self.env.extend(d.var.name, c())
        return d.var.name

    def visit_assigner(self, a: Assigner) -> values.Value:
        n: str = a.decl.accept(self)
        return self.env.set_val(n, a.value.value)

    def visit_trigger(self, t: Trigger):
        # TODO: source should be stored in environment
        self.events[t.var1.name].append(lambda: t.math_funcs.accept(self))

    def visit_plotter(self, pltr: Plotter):
        graph: ConcreteGraphType = pltr.graph.accept(self)
        line: bool = isinstance(graph, LineXYGraph)

        def add_data(m: GraphManager):
            m.add_plot(pltr.graph_name, line_plot=line)
        self.do_graphics(add_data)

        self.plots.append((pltr.graph_name, pltr.x, pltr.y))

    def visit_var(self, v: Var):
        return self.env.get_val(v.name)

    def visit_source(self, s: Source) -> str:
        return s.url

    def visit_type(self, t: Type):
        return t.type

    def visit_value(self, v: Value) -> values.Value:
        return v.value

    def visit_graph(self, g: Graph) -> ConcreteGraphType:
        return g.graph

    def visit_var_axis(self, va: VarAxis) -> values.Value:
        return va.var.accept(self)

    def visit_func_axis(self, fa: FuncAxis):
        return fa.fun.accept(self)

    def visit_math_funcs(self, mf: MathFuncs):
        for func in mf.mth_func_lst:
            func.accept(self)
            pass
        pass

    def visit_simple_func(self, smp: SimpleFunc) -> values.Value:
        n: str = smp.impacted_var.name
        if isinstance(smp.rhs.value, values.NumericalValue):
            v: values.Value = self.env.get_val(n)
            if isinstance(v, values.NumericalValue):
                operand = smp.op.accept(self)
                result: values.NumericalValue = apply_op(operand,v, smp.rhs.value)
                return self.env.set_val(n, result)
            else:
                raise LanguageError("Simple Functions only accept numbers.")
        else:
            raise LanguageError("Simple Functions only accept numbers.")

    def visit_increment(self, i: Increment) -> values.Value:
        n = i.impacted_var.name
        v: values.Value = self.env.get_val(n)
        if isinstance(v, values.NumericalValue):
            result: values.NumericalValue = apply_qk("inc", v)
            return self.env.set_val(n, result)
        else:
            raise LanguageError("Fast Functions only accept numbers.")

    def visit_decrement(self, d: Decrement) -> values.Value:
        n = d.impacted_var.name
        v: values.Value = self.env.get_val(n)
        if isinstance(v, values.NumericalValue):
            result: values.NumericalValue = apply_qk("dec", v)
            return self.env.set_val(n, result)
        else:
            raise LanguageError("Fast Functions only accept numbers.")

    def visit_builtin_func(self, bltn: BuiltinFunc):
        v: values.Value = self.env.get_val(bltn.input.name)
        if isinstance(v, values.NumericalValue):
            return apply_fn(bltn.fun, v)
        else:
            raise LanguageError("BuiltIn Functions only accept numbers.")

    def visit_operand(self, op: Operand) -> numerical.NumOp:
        return op.op



from typing import Optional, cast


from mystix.language.shared import ast
from mystix.language.tokenization import Tokenizer
from mystix.language.shared.primitives import Types
from mystix.language.shared.primitives import values, graphs
from mystix.language.shared.primitives.numerical import NumOp, NumFunction


class ParseError(Exception):
    pass


class Parser:

    operators = {
        '+=': NumOp.PLUS,
        '-=': NumOp.MINUS,
        '*=': NumOp.TIMES,
        '/=': NumOp.DIV,
        '^=': NumOp.EXP
    }
    bltn = {
        'log': NumFunction.LOG,
        'sin': NumFunction.SIN,
        'cos': NumFunction.COS,
        'exp': NumFunction.EXP
    }

    def __init__(self, tokenizer: Tokenizer):
        self.tokenizer = tokenizer

    def parseProgram(self) -> ast.Program:
        commands = []
        self.tokenizer.get_and_check_next("program:")
        self.tokenizer.get_and_check_next(";")
        while(self.tokenizer.more_tokens() and self.tokenizer.check_next() != "start!"):
            command = self.parseCommand()
            commands.append(command)
        self.tokenizer.get_and_check_next("start!")
        return ast.Program(ast.Body(commands))
    
    def parseCommand(self) -> ast.Command:
        next_token = self.tokenizer.check_next()
        command: Optional[ast.Command] = None
        if (next_token == "map"):
            #print("Parsing Map")
            command = self.parseMapper()
        elif (next_token == "observe"):
            #print("Parsing Trigger")
            command = self.parseTrigger()
        elif (next_token == "plot"):
            #print("Parsing Plotter")
            command = self.parsePlotter()
        # Line is either Loader or Assigner
        line = self.tokenizer.get_line()
        if ("remote" in line):
            #print('Parsing Loader')
            command = self.parseLoader()
        elif ("=" in line and "remote" not in line):
            #print('Parsing Assigner')
            command = self.parseAssigner()
        while (self.tokenizer.check_next() in ['\n', ';']):
            self.tokenizer.get_next()
        if command is not None:
            return command
        else:
            raise ParseError("Could not determine command")
    
    def parseLoader(self) -> ast.Loader:
        var = self.parseVar()
        self.tokenizer.get_and_check_next("=")
        source = self.parseSource()
        return ast.Loader(var, source)

    def parseMapper(self) -> ast.Mapper:
        self.tokenizer.get_and_check_next("map")
        self.tokenizer.get_and_check_next("\(")
        var = ast.Var(self.tokenizer.get_next())
        self.tokenizer.get_and_check_next("\)")
        map_from = self.parseString()
        self.tokenizer.get_and_check_next("to")
        declare = self.parseDeclare()
        return ast.Mapper(var, map_from, declare)
    
    def parseDeclare(self) -> ast.Declare:
        var_type = self.parseType()
        var = self.parseVar()
        return ast.Declare(var_type, var)

    def parseAssigner(self) -> ast.Assigner:
        declare = self.parseDeclare()
        self.tokenizer.get_and_check_next("=")
        value = self.parseValue()
        return ast.Assigner(declare, value)

    def parseTrigger(self) -> ast.Trigger:
        self.tokenizer.get_and_check_next("observe")
        self.tokenizer.get_and_check_next("\(")
        var = self.parseVar()
        self.tokenizer.get_and_check_next("\)")
        self.tokenizer.get_and_check_next("do")
        math_funcs = self.parseMathFuncs()
        return ast.Trigger(var, math_funcs)

    def parsePlotter(self) -> ast.Plotter:
        self.tokenizer.get_and_check_next("plot")
        graph = self.parseGraph()
        self.tokenizer.get_and_check_next("\(")
        x_axis = self.parseAxis(',')
        #print("X AXIS GOTTEN")
        self.tokenizer.get_and_check_next(',')
        #print("COMMA CONFIRMED")
        y_axis = self.parseAxis(')')
        self.tokenizer.get_and_check_next("\)")
        self.tokenizer.get_and_check_next("titled")
        name = self.parseString()
        return ast.Plotter(graph, x_axis, y_axis, name)

    def parseSource(self) -> ast.Source:
        self.tokenizer.get_and_check_next("remote")
        self.tokenizer.get_and_check_next("\(")
        self.tokenizer.get_and_check_next("\"")
        url = self.tokenizer.get_next()
        self.tokenizer.get_and_check_next("\"")
        self.tokenizer.get_and_check_next("\)")
        return ast.Source(url)

    def parseMathFuncs(self) -> ast.MathFuncs:
        functions = []
        line = self.tokenizer.get_line(',')
        #print(line)
        functions.append(self.parseFunc(line))
        next_token = self.tokenizer.get_next()
        while(next_token not in ['\n', ';'] and next_token == ','):
            line = self.tokenizer.get_line(',')
            functions.append(self.parseFunc(line))
            next_token = self.tokenizer.get_next()
        return ast.MathFuncs(functions)
    
    def parseAxis(self, endline = ',') -> ast.Axis:
        line = self.tokenizer.get_line(endline)
        #print(line)
        if(len(line) == 1):
            return ast.VarAxis(ast.Var(self.tokenizer.get_next()))
        elif self.isMathFunc(line):
            func = self.parseFunc(line)
            return ast.FuncAxis(func)
        else:
            raise ParseError("Could not parse axis")

    def parseFunc(self, line) -> ast.Func:
        func: Optional[ast.Func] = None
        if (self.isFastFunc(line)):
            func = self.parseFastFunc()
        elif (self.isSimpFunc(line)):
            func = self.parseSimpFunc()
        elif (self.isBltnFunc(line)):
            func = self.parseBltnFunc()
        #print(func)
        #print(self.tokenizer.check_next())
        if func is not None:
            return func
        else:
            raise ParseError("Could not determine function")

    def parseSimpFunc(self) -> ast.SimpleFunc:
        var = ast.Var(self.tokenizer.get_next())
        op = self.tokenizer.get_next()
        op += self.tokenizer.get_next()
        concrete_op: NumOp = self.operators[op]
        value = self.parseValue()
        return ast.SimpleFunc(var, ast.Operand(concrete_op), value)
    
    def parseFastFunc(self) -> ast.FastFunc:
        token = self.tokenizer.get_next()
        var = ast.Var(token[:-2])
        operator = token[-2:]
        if operator == '++':
            return ast.Increment(var)
        else:
            return ast.Decrement(var)
    
    def parseBltnFunc(self) -> ast.BuiltinFunc:
        op = self.tokenizer.get_next()
        f: NumFunction = self.bltn[op]
        self.tokenizer.get_and_check_next("\(")
        var = self.tokenizer.get_next()
        self.tokenizer.get_and_check_next("\)")
        return ast.BuiltinFunc(f, op)

    def parseString(self) -> str:
        self.tokenizer.get_and_check_next('"')
        string_token = ''
        while(self.tokenizer.check_next() != '"'):
            string_token += self.tokenizer.get_next()
        self.tokenizer.get_and_check_next('"')
        return string_token

    def parseGraph(self) -> ast.Graph:
        graph = self.tokenizer.get_next()
        if graph == 'scatter_xy':
            return ast.Graph(graphs.ScatterXYGraph())
        elif graph == 'line_xy':
            return ast.Graph(graphs.LineXYGraph())
        return ast.Graph(graph)

    def parseVar(self) -> ast.Var:
        return ast.Var(self.tokenizer.get_next())

    def parseValue(self) -> ast.Value:
        next_token = self.tokenizer.get_next()
        val = next_token
        # String Case
        if next_token == '"':
            val = ''
            next_token = self.tokenizer.get_next()
            val += next_token
            while (next_token != '"'):
                val += self.tokenizer.get_next()
            return ast.Value(values.CategoricalValue(val))
        # Binary Boolean case
        elif next_token.lower() == 'true':
            return ast.Value(values.BinaryValue(True))
        elif next_token.lower() == 'false':
            return ast.Value(values.BinaryValue(False))
        else:
            try:
                if "." in val:
                    return ast.Value(values.FloatValue(float(val)))
                else:
                    return ast.Value(values.IntegerValue(int(val)))
            except TypeError:
                raise ParseError("Could not parse value '%s'" % val)

    def parseType(self) -> ast.Type:
        t = self.tokenizer.get_next()
        if t == 'number':
            return ast.Type(Types.NUMBER)
        elif t == 'binary':
            return ast.Type(Types.BINARY)
        else:
            return ast.Type(Types.CATEGORY)

    def isSep(self, input: str) -> bool:
        if (input == ';' or input == '\n'):
            return True
        return False

    def isMathFunc(self, line):
        if (self.isSimpFunc(line) or self.isFastFunc(line) or self.isBltnFunc(line)):
            return True
        return False

    def isFastFunc(self, line):
        ops = ['++', '--']
        if (len(line) == 1 and line[0][-2:] in ops):
            return True
        return False

    def isSimpFunc(self, line):
        if (len(line) == 4 and str(line[1] + line[2]) in self.operators):
            return True
        return False

    def isBltnFunc(self, line):
        if (line[0] in self.bltn):
            return True
        return False



from .Node import Node
from mystix.language.shared.ast.visitor import Visitor


class Var(Node):

    def __init__(self, name: str):
        self.name = name

    def accept(self, v: Visitor):
        return v.visit_var(self)



from mystix.language.shared.ast import *
from typing import cast
from mystix.util.errors import NonExhaustiveTypeCaseError


def value_equal(v1: Value, v2: Value) -> bool:
    return v1.value.equals(v2.value)


def type_equal(t1: Type, t2: Type) -> bool:
    return t1.type == t2.type


def var_equal(v1: Var, v2: Var) -> bool:
    return v1.name == v2.name


def declare_equal(d1: Declare, d2: Declare) -> bool:
    return type_equal(d1.type, d2.type) and var_equal(d1.var, d2.var)


def operand_equal(o1: Operand, o2: Operand) -> bool:
    return o1.op == o2.op


def builtin_func_equal(f1: BuiltinFunc, f2: BuiltinFunc) -> bool:
    return f1.fun == f2.fun and var_equal(f1.input, f2.input)


def fast_func_equal(f1: FastFunc, f2: FastFunc) -> bool:
    t = type(f1)
    if t is not type(f2):
        return False
    else:
        return var_equal(f1.impacted_var, f2.impacted_var)


def simple_func_equal(f1: SimpleFunc, f2: SimpleFunc) -> bool:
    return var_equal(f1.impacted_var, f2.impacted_var) \
           and operand_equal(f1.op,f2.op) \
           and value_equal(f1.rhs, f2.rhs)


def func_equal(f1: Func, f2: Func) -> bool:
    t = type(f1)
    if t is not type(f2):
        return False
    elif t is BuiltinFunc:
        return builtin_func_equal(cast(BuiltinFunc, f1), cast(BuiltinFunc, f2))
    elif isinstance(f1, FastFunc):
        return fast_func_equal(cast(FastFunc, f1), cast(FastFunc, f2))
    elif t is SimpleFunc:
        return simple_func_equal(cast(SimpleFunc, f1), cast(SimpleFunc, f2))
    else:
        raise NonExhaustiveTypeCaseError()


def axis_equal(a1: Axis, a2: Axis) -> bool:
    t = type(a1)
    if t is not type(a2):
        return False
    elif t is VarAxis:
        return var_equal(cast(VarAxis, a1).var, cast(VarAxis, a2).var)
    elif t is FuncAxis:
        return func_equal(cast(FuncAxis, a1).fun, cast(FuncAxis, a2).fun)
    else:
        raise NonExhaustiveTypeCaseError()


def graph_equal(g1: Graph, g2: Graph) -> bool:
    return g1.graph.equals(g2.graph)


def plotter_equal(p1: Plotter, p2: Plotter) -> bool:
    return graph_equal(p1.graph, p2.graph)


def mapper_equal(m1: Mapper, m2: Mapper) -> bool:
    return var_equal(m1.src, m2.src) and m1.tbl_field == m2.tbl_field and \
        declare_equal(m1.decl, m2.decl)


def source_equal(s1: Source, s2: Source) -> bool:
    return s1.url == s2.url


def loader_equal(l1: Loader, l2: Loader) -> bool:
    return var_equal(l1.name, l2.name) and source_equal(l1.source, l2.source)


def assigner_equal(a1: Assigner, a2: Assigner) -> bool:
    return declare_equal(a1.decl, a2.decl) and value_equal(a1.value, a2.value)


def math_funcs_equal(fs1: MathFuncs, fs2: MathFuncs) -> bool:
    l1 = len(fs1.mth_func_lst)
    if l1 != len(fs2.mth_func_lst):
        return False
    else:
        for i in range(l1):
            if not func_equal(fs1.mth_func_lst[i], fs2.mth_func_lst[i]):
                return False
        return True


def trigger_equal(t1: Trigger, t2: Trigger) -> bool:
    return var_equal(t1.var1, t2.var1) \
           and math_funcs_equal(t1.math_funcs, t2.math_funcs)


def command_equal(c1: Command, c2: Command) -> bool:
    t = type(c1)
    if t is not type(c2):
        return False
    elif t is Command:
        return True
    elif t is Assigner:
        return assigner_equal(cast(Assigner, c1), cast(Assigner, c2))
    elif t is Loader:
        return loader_equal(cast(Loader, c1), cast(Loader, c2))
    elif t is Mapper:
        return mapper_equal(cast(Mapper, c1), cast(Mapper, c2))
    elif t is Plotter:
        return plotter_equal(cast(Plotter, c1), cast(Plotter, c2))
    elif t is Trigger:
        return trigger_equal(cast(Trigger, c1), cast(Trigger, c2))
    else:
        raise NonExhaustiveTypeCaseError()


def body_equal(b1: Body, b2: Body) -> bool:
    l1 = len(b1.commands)
    if l1 != len(b2.commands):
        return False
    else:
        for i in range(l1):
            if not command_equal(b1.commands[i], b2.commands[i]):
                return False
        return True


def ast_equal(p1: Program, p2: Program) -> bool:
    """
    Should return true if p1 and p2 represent the same program
    """
    return body_equal(p1.body, p2.body)  # stub



"""
AST Data Structures
This is intended to reflect the grammar as defined in the grammar.txt document. Any
change in one definition should accompany a change in the other so that we maintain
consistency.
"""

from .Node import Node
from .Program import Program
from .Body import Body
from .commands_ast.Command import Command
from .commands_ast.Loader import Loader
from .commands_ast.Mapper import Mapper
from .commands_ast.Plotter import Plotter
from .commands_ast.Trigger import Trigger
from .commands_ast.Assigner import Assigner
from .Var import Var
from .Source import Source
from .Type import Type
from .Value import Value
from .graphs_ast.Graph import Graph
from .graphs_ast.Axis import Axis
from .graphs_ast.Axis import VarAxis, FuncAxis
from .math_ast.MathFuncs import MathFuncs
from .math_ast.Func import Func
from .math_ast.SimpleFunc import SimpleFunc
from .math_ast.FastFunc import FastFunc, Increment, Decrement
from .math_ast.BuiltinFunc import BuiltinFunc
from .math_ast.Operand import Operand
from .Declare import Declare



"""
Contains all internal data-structures which represent
concrete items in our engine. These are expressed with
the 'Concrete' prefix so as to not pollute the abstract
syntax namespace.
"""

from .types import Types
from .values import Value as ConcreteValue
from .graphs import Graph as ConcreteGraph
from .numerical import NumOp as ConcreteNumOp



import sys
import unittest

from mypy import api


class TypingTests(unittest.TestCase):

    def test_types(self):
        result = api.run(["mystix", "tests"])
        if result[0]:
            print('\nType checking report:\n')
            print(result[0])  # stdout
        if result[1]:
            print('\nError report:\n')
            print(result[1])  # stderr
        status: int = result[2]
        if result[2] != 0:
            self.fail(status)
        print('\nExit status:', status)



from ..Node import Node
from mystix.language.shared.primitives import ConcreteNumOp
from mystix.language.shared.ast.visitor import Visitor


class Operand(Node):

    def __init__(self, o: ConcreteNumOp):
        self.op = o
        pass

    def accept(self, v: Visitor):
        return v.visit_operand(self)



from typing import Callable


class FileLoadError(Exception):
    pass


def read_program_file(filename: str, f: Callable):
    """
    Read program content and execute f as a procedure with
    content as input
    """
    result = None
    with open(filename, "r") as file:
        s: str = file.read()
        result = f(s)
    return result



from ..Node import Node
from ..visitor import Visitor


class Command(Node):
    """
    Parent class to Loader, Mapper, Holder, Math, and Plotter
    We assume all visitors visit the subclasses, but provide
    the accept stub for static analysis.
    """

    def accept(self, v: Visitor):
        pass



from unittest import TestCase
from mystix.language.shared.primitives.values import (
    NumericalValue, FloatValue, IntegerValue )
from mystix.language.shared.primitives.numerical import NumOp, NumFunction
from mystix.targets.analysis.math import apply_fn, apply_op, MathError
from typing import cast


from math import pi


class MathTests(TestCase):

    def test_plus(self):
        result: NumericalValue = apply_op(
            NumOp.PLUS,
            cast(NumericalValue, IntegerValue(1)),
            cast(NumericalValue, IntegerValue(2)))
        self.assertTrue(type(result) is IntegerValue)
        self.assertTrue(result.equals(IntegerValue(3)))

    def test_minus(self):
        result: NumericalValue = apply_op(
            NumOp.MINUS,
            cast(NumericalValue, IntegerValue(1)),
            cast(NumericalValue, IntegerValue(2)))
        self.assertTrue(type(result) is IntegerValue)
        self.assertTrue(result.equals(IntegerValue(-1)))

    def test_plus_with_float(self):
        result: NumericalValue = apply_op(
            NumOp.PLUS,
            cast(NumericalValue, IntegerValue(1)),
            cast(NumericalValue, FloatValue(2.0)))
        self.assertTrue(type(result) is FloatValue)
        self.assertTrue(result.equals(FloatValue(3.0)))

    def test_divide_zero(self):
        try:
            apply_op(
                NumOp.DIV,
                cast(NumericalValue, IntegerValue(1)),
                cast(NumericalValue, IntegerValue(0)))
            self.fail()
        except MathError:
            pass

    def test_pow(self):
        result: NumericalValue = apply_op(
            NumOp.EXP,
            cast(NumericalValue, IntegerValue(2)),
            cast(NumericalValue, IntegerValue(3)))
        self.assertTrue(type(result) is IntegerValue)
        self.assertTrue(result.equals(IntegerValue(8)))

    def test_sin(self):
        result: FloatValue = apply_fn(
            NumFunction.SIN,
            cast(NumericalValue, FloatValue(pi/2)))
        self.assertTrue(result.equals(FloatValue(1.0)))

    def test_log_domain(self):
        try:
            apply_fn(
                NumFunction.LOG,
                cast(NumericalValue, IntegerValue(-1)))
            self.fail()
        except MathError:
            pass



import unittest
from typing import cast

from mystix.language.evaluation.vars import Environment, UndefinedVariableError
from mystix.language.shared.primitives.values import IntegerValue, Value


class EnvironmentTests(unittest.TestCase):

    def test_undefined(self):
        e = Environment()
        try:
            e.get_val("color")
            self.fail()
        except UndefinedVariableError:
            pass

    def test_undefined_undef(self):
        e = Environment()
        try:
            e.undef("color")
            self.fail()
        except UndefinedVariableError:
            pass

    def test_set_no_extend(self):
        e = Environment()
        try:
            e.set_val("width", IntegerValue(5))
            self.fail()
        except UndefinedVariableError:
            pass

    def test_basic(self):
        e = Environment()
        e.extend("width", IntegerValue(3))
        e.extend("height", IntegerValue(4))
        self.assertEqual(3, cast(IntegerValue, e.get_val("width")).value)
        self.assertEqual(4, cast(IntegerValue, e.get_val("height")).value)

    def test_basic_set(self):
        e = Environment()
        e.extend("width", IntegerValue(3))
        self.assertEqual(3, cast(IntegerValue, e.get_val("width")).value)
        e.set_val("width", IntegerValue(8))
        self.assertEqual(8, cast(IntegerValue, e.get_val("width")).value)

    def test_undef(self):
        e = Environment()
        e.extend("width", IntegerValue(3))
        self.assertEqual(3, cast(IntegerValue, e.get_val("width")).value)
        e.undef("width")
        try:
            e.get_val("width")
            self.fail()
        except UndefinedVariableError:
            pass

    def test_multiple_extend(self):
        e = Environment()
        e.extend("width", IntegerValue(3))
        e.extend("width", IntegerValue(6))
        self.assertEqual(6, cast(IntegerValue, e.get_val("width")).value)

    def test_multiple_extend_undef(self):
        e = Environment()
        e.extend("width", IntegerValue(3))
        e.extend("width", IntegerValue(6))
        self.assertEqual(6, cast(IntegerValue, e.get_val("width")).value)
        e.undef("width")
        self.assertEqual(3, cast(IntegerValue, e.get_val("width")).value)



from unittest import TestCase
from tests.util.print_program import ProgramPrinter
from tests.util.example_ast import example_1


class VisitorInterfaceTests(TestCase):

    def test_print_visitor(self):
        self.assertEqual(
            "(Program "
                "(Body "
                    "(Loader "
                    "(Var 'source') "
                    "(Source "
                        "'www.coviddata.com/stream')) "
                    "(Mapper "
                        "(Var 'source') "
                        "'case_date' "
                        "(Declare (Type Types.NUMBER) (Var 'date'))) "
                    "(Assigner "
                        "(Declare (Type Types.NUMBER) (Var 'count')) "
                        "(Value 0)) "
                    "(Trigger (Var 'source') ((Increment (Var 'count')))) "
                    "(Plotter (Graph 'scatter') (VarAxis (Var 'date')) "
                        "(VarAxis (Var 'age')) 'age_graph') "
                    "(Plotter (Graph 'line') (VarAxis (Var 'date')) "
                        "(FuncAxis (BuiltinFunc NumFunction.LOG (Var 'count'))) "
            "'age_graph')))",
            ProgramPrinter().print(example_1()))



from unittest import TestCase
from typing import cast
from mystix.language.evaluation import Evaluator
from mystix.language.shared import ast
from mystix.targets.data.dataLoader import DataLoaderError
from mystix.language.shared.primitives import Types, ConcreteNumOp
from mystix.language.shared.primitives.values import IntegerValue, FloatValue


class LoaderEvaluationTests(TestCase):

    def test_duplicate_name(self):
        e = Evaluator()
        p = ast.Program(
            ast.Body([
                ast.Loader(ast.Var("source"),
                           ast.Source("www.source.com")),
                ast.Loader(ast.Var("source"),
                           ast.Source("www.source2.com")),
            ]))
        code, err = e.evaluate(p, duration=4000)
        self.assertNotEqual(0, code)
        self.assertTrue(isinstance(err, DataLoaderError))

    def test_invalid_host_no_trigger(self):
        e = Evaluator()
        p = ast.Program(
            ast.Body([
                ast.Loader(ast.Var("source"),
                           ast.Source("https://covid-apiiii.com/api/reports")),
            ]))
        code, err = e.evaluate(p, duration=4000)
        self.assertNotEqual(0, code)
        self.assertTrue(isinstance(err, DataLoaderError))

    def test_invalid_host(self):
        e = Evaluator()
        p = ast.Program(
            ast.Body([
                ast.Loader(ast.Var("source"),
                           ast.Source("https://covid-apiiii.com/api/reports")),
                ast.Trigger(ast.Var("source"), ast.MathFuncs([ast.Increment(ast.Var(
                    "count"))])),
            ]))
        code, err = e.evaluate(p, duration=4000)
        self.assertNotEqual(0, code)
        self.assertTrue(isinstance(err, DataLoaderError))

    def test_valid_source(self):
        e = Evaluator()
        p = ast.Program(
            ast.Body([
                ast.Loader(ast.Var("source"),
                           ast.Source("https://covid-api.com/api/reports")),
                ast.Assigner(ast.Declare(ast.Type(Types.NUMBER), ast.Var("count")),
                             ast.Value(IntegerValue(0))),
                ast.Trigger(ast.Var("source"), ast.MathFuncs([ast.Increment(ast.Var(
                    "count"))])),
            ]))
        code, err = e.evaluate(p, duration=4000)
        self.assertEqual(0, code)


class MiscEvaluationTests(TestCase):

    def test_simple_func_baseline(self):
        e = Evaluator()
        p = ast.Program(
            ast.Body([
                ast.Loader(ast.Var("source"),
                           ast.Source("https://covid-api.com/api/reports")),
                ast.Assigner(ast.Declare(ast.Type(Types.NUMBER), ast.Var("count")),
                             ast.Value(IntegerValue(20))),
                ast.Trigger(ast.Var("source"), ast.MathFuncs([])),
            ]))
        code, err = e.evaluate(p, duration=7000)
        self.assertEqual(0, code)
        self.assertEqual(20, cast(IntegerValue, e.env.get_val("count")).value)

    def test_simple_func_divide(self):
        e = Evaluator()
        p = ast.Program(
            ast.Body([
                ast.Loader(ast.Var("source"),
                           ast.Source("https://covid-api.com/api/reports")),
                ast.Assigner(ast.Declare(ast.Type(Types.NUMBER), ast.Var("count")),
                             ast.Value(IntegerValue(20))),
                ast.Trigger(ast.Var("source"), ast.MathFuncs([ast.SimpleFunc(
                    ast.Var("count"), ast.Operand(ConcreteNumOp.DIV),
                    ast.Value(IntegerValue(2)))])),
            ]))
        code, err = e.evaluate(p, duration=7000)
        self.assertEqual(0, code)
        self.assertTrue(cast(IntegerValue, e.env.get_val("count")).value < 20)

    def test_simple_func_pow(self):
        e = Evaluator()
        p = ast.Program(
            ast.Body([
                ast.Loader(ast.Var("source"),
                           ast.Source("https://covid-api.com/api/reports")),
                ast.Assigner(ast.Declare(ast.Type(Types.NUMBER), ast.Var("count")),
                             ast.Value(FloatValue(1.1))),
                ast.Trigger(ast.Var("source"), ast.MathFuncs([ast.SimpleFunc(
                    ast.Var("count"), ast.Operand(ConcreteNumOp.EXP),
                    ast.Value(IntegerValue(2)))])),
            ]))
        code, err = e.evaluate(p, duration=7000)
        self.assertEqual(0, code)
        self.assertTrue(cast(IntegerValue, e.env.get_val("count")).value > 1)



import csv, urllib.request, requests, json  # type: ignore
from typing import List

from mystix.language.evaluation.errors import LanguageError


class DataLoaderError(LanguageError):
    pass


class MockDateGenerator:

    def __init__(self):
        self.year = 2020
        self.month = 3
        self.day = 1

    def get(self) -> str:
        date = "%d-%02d-%02d" % (self.year, self.month, self.day)
        self.day = self.day + 1
        if self.day > 28:
            self.month = self.month + 1
            self.day = 1
        if self.month > 12:
            self.year = self.year + 1
            self.month = 1
        return date


class DataSource:

    """
    Encapsulates a data source that updates over time
    Current implementation is limited to supporting http
    URLs.
    We are also still limited in the format of data received -
    namely, an error will be thrown if the responses returned
    by the source do not match a preset format (e.g. is valid JSON
    and contains a 'data' field)
    """

    def __init__(self, url: str):
        self.url = url
        self.dates = MockDateGenerator()

    def get_new(self) -> List:
        """
        Returns a list of JSON-like dicts whose field names
        correspond to mappable headers
        Data format expectation is hardcoded until we introduce
        features to match data appropriately
        simulates a stream by getting data for new dates each time
        """
        d: str = self.dates.get()
        try:
            req = requests.get(self.url, params={'date': d})
        except requests.exceptions.MissingSchema:
            raise DataLoaderError("Source URL %s is invalid" % self.url)
        except requests.exceptions.ConnectionError:
            raise DataLoaderError("Could not reach host %s" % self.url)
        res = req.text
        if req.status_code != 200:
            raise DataLoaderError("Remote responded with error invalid code")
        # with open("tmp/load_cache/api_response_%s.json" % d, "w") as cache:
        #     cache.write(res)
        try:
            obj = json.loads(res)
        except json.decoder.JSONDecodeError:
            raise DataLoaderError("Remote source produced non-json")
        if 'data' not in obj or type(obj['data']) is not list:
            raise DataLoaderError("New data did not match expected format")
        # hard-coded filter TODO should add this as a feature
        return [r for r in obj['data'] if r['region']['iso'] == "CAN"]


class DataLoader:

    def __init__(self):
        self.sources = {}

    def register_source(self, url: str, src_id: str):
        if src_id in self.sources:
            raise DataLoaderError("Source '%s' is already registered" % src_id)
        else:
            self.sources[src_id] = DataSource(url)

    def get_new(self, src_id: str) -> List:
        if src_id not in self.sources:
            raise DataLoaderError("Source '%s' does not exist" % src_id)
        else:
            s: DataSource = self.sources[src_id]
            return s.get_new()


# def load_data(url: str, file: str = None):
#     response = urllib.request.urlopen(url)
#     lines = [l.decode('utf-8') for l in response.readlines()]
#     data = csv.reader(lines)
#
#     if file is not None:
#         text_file = open(file, "w")
#
#     for row in data:
#         if file is not None:
#             text_file.write("[")
#             for s in row:
#                 text_file.write("'"+s+"', ")
#             text_file.write("]\n")
#
#     if file is not None:
#         text_file.close()
#
#     return data



from mystix.language.shared.ast.visitor import Visitor
from abc import ABC, abstractmethod


class Node(ABC):
    """
    The highest-level type in the syntax tree. Should not be directly
    instantiated.
    """
    
    @abstractmethod
    def accept(self, v: Visitor):
        pass



from mystix.targets.visualization.graphs import GraphManager
from threading import Thread
import time
from math import sin, pi
import numpy as np  # type: ignore

MOCK_DATA_RESOLUTION = 50


# MOCK_DATA_X = [t for t in np.linspace(0, 2*pi, MOCK_DATA_RESOLUTION)]
# MOCK_DATA_Y = [sin(t) for t in MOCK_DATA_X]
# mock_ptr = 0
# def get_mock_data(count: int):
#     x_data = []
#     y_data = []
#     global mock_ptr
#     for i in range(count):
#         x_data.append(
#             MOCK_DATA_X[
#                 (mock_ptr + i) % len(MOCK_DATA_X)] + moc)
#         y_data.append(
#             MOCK_DATA_Y[
#                 (mock_ptr + i) % len(MOCK_DATA_Y)])
#     mock_ptr = mock_ptr + count
#     return x_data, y_data

t = 0.0


def get_mock_data(count: int):
    x_data = []
    y_data = []
    global t
    for i in range(count):
        x_data.append(t)
        y_data.append(sin(t))
        t = t + (2*pi/MOCK_DATA_RESOLUTION)
    return x_data, y_data


def run():
    gm = GraphManager()
    gm.add_plot("example_plot", line_plot=True)
    def run_mock_logic():
        for i in range(30): # 3 seconds of playback
            t,y = get_mock_data(5)
            gm.add_plot_data("example_plot", t, y)
            time.sleep(0.1)

    logic = Thread(target=run_mock_logic)
    logic.start()

    gm.graphics.display(ttl=3000)

    logic.join()
    gm.clean()



from ..Node import Node
from mystix.language.shared.primitives import ConcreteGraph
from mystix.language.shared.ast.visitor import Visitor


class Graph(Node):

    def __init__(self, g: ConcreteGraph):
        self.graph = g

    def accept(self, v: Visitor):
        return v.visit_graph(self)



from .Func import Func
from ..Var import Var
from mystix.language.shared.primitives.numerical import NumFunction
from mystix.language.shared.ast.visitor import Visitor


class BuiltinFunc(Func):

    def __init__(self, fun: NumFunction, x: Var):
        self.fun: NumFunction = fun
        self.input: Var = x

    def accept(self, v: Visitor):
        return v.visit_builtin_func(self)




import unittest as ut
from mystix.ui.util import read_program_file
from mystix.language.tokenization.tokenizer import Tokenizer, TokenizationError
from mystix.language.parsing.parser import Parser
from mystix.language.shared.ast import *
from mystix.language.shared.primitives import Types
from mystix.language.shared.primitives import values
from mystix.language.shared.primitives.graphs import LineXYGraph, ScatterXYGraph
from mystix.language.shared.primitives.numerical import NumOp, NumFunction


class TestParse(ut.TestCase):

    def test_program(self):
        tokens, parser = self.setup("parse_program")
        res = parser.parseProgram()
        commands = res.body.commands
        self.assertIsInstance(commands[0].graph.graph, LineXYGraph)
        self.assertEqual(commands[0].x.var.name, 'x')
        self.assertEqual(commands[0].y.fun.fun, NumFunction.LOG)
        self.assertEqual(commands[0].graph_name, 'line')
        self.assertEqual(commands[1].var1.name, 'triggerVar')
        self.assertEqual(len(commands[1].math_funcs.mth_func_lst), 2)
        self.assertEqual(commands[1].math_funcs.mth_func_lst[0].impacted_var.name, 'triggerVar')
        self.assertEqual(commands[1].math_funcs.mth_func_lst[1].impacted_var.name, 'triggerVar')
        self.assertEqual(commands[1].math_funcs.mth_func_lst[1].op.op, NumOp.PLUS)
        self.assertIsInstance(commands[1].math_funcs.mth_func_lst[1].rhs.value,
                              values.IntegerValue)
        self.assertTrue(commands[1].math_funcs.mth_func_lst[1].rhs.value.equals(
            values.IntegerValue(1)))
        self.assertEqual(commands[2].decl.type.type, Types.CATEGORY)
        self.assertEqual(commands[2].decl.var.name, 'A')
        self.assertTrue(commands[2].value.value.equals(values.IntegerValue(123456)))

    def test_example1(self):
        tokens, parser = self.setup("example1")
        res = parser.parseProgram()
        print("Parsed")
        commands = res.body.commands
        self.assertIsInstance(commands[0], Loader)
        self.assertIsInstance(commands[1], Mapper)
        self.assertIsInstance(commands[2], Assigner)
        self.assertIsInstance(commands[3], Trigger)
        self.assertIsInstance(commands[4], Plotter)
        self.assertIsInstance(commands[5], Plotter)

    def test_loader(self):
        tokens, parser = self.setup("parse_loader")
        res = parser.parseLoader()
        self.assertEqual(res.name.name, "x")
        self.assertEqual(res.source.url, "www.coviddata.com/stream")
    
    def test_fast_func(self):
        tokens, parser = self.setup("parse_fast_func")
        res = parser.parseFastFunc()
        self.assertEqual(res.impacted_var.name, 'x')
        self.assertIsInstance(res, Increment)

    def test_simp_func(self):
        tokens, parser = self.setup("parse_simp_func")
        res = parser.parseFunc(tokens)
        self.assertEqual(res.impacted_var.name, 'x')
        self.assertEqual(res.op.op, NumOp.PLUS)
        self.assertIsInstance(res, SimpleFunc)
    
    def test_mapper(self):
        tokens, parser = self.setup("parse_mapper")
        res = parser.parseMapper()
        self.assertEqual(res.src.name, 'testVar')
        self.assertEqual(res.tbl_field, 'testingAxis')
        self.assertEqual(res.decl.type.type, Types.CATEGORY)
        self.assertEqual(res.decl.var.name, 'A')

    def test_assigner(self):
        tokens, parser = self.setup("parse_assigner")
        res = parser.parseAssigner()
        self.assertEqual(res.decl.type.type, Types.CATEGORY)
        self.assertEqual(res.decl.var.name, 'A')
        self.assertEqual(res.value.value.value, 123456)
    
    def test_math_funcs(self):
        tokens, parser = self.setup("parse_math_funcs")
        res = parser.parseMathFuncs()
        self.assertEqual(len(res.mth_func_lst), 3)
        self.assertIsInstance(res.mth_func_lst[0], Increment)
        self.assertIsInstance(res.mth_func_lst[1], SimpleFunc)
        self.assertIsInstance(res.mth_func_lst[2], BuiltinFunc)

    def test_trigger(self):
        tokens, parser = self.setup("parse_trigger")
        res = parser.parseTrigger()
        self.assertEqual(res.var1.name, 'triggerVar')
        self.assertEqual(len(res.math_funcs.mth_func_lst), 2)
        self.assertEqual(res.math_funcs.mth_func_lst[0].impacted_var.name, 'triggerVar')
        self.assertEqual(res.math_funcs.mth_func_lst[1].impacted_var.name, 'triggerVar')
        self.assertEqual(res.math_funcs.mth_func_lst[1].op.op, NumOp.PLUS)
        self.assertTrue(res.math_funcs.mth_func_lst[1].rhs.value.equals(
            values.IntegerValue(1)))
    
    def test_plotter(self):
        tokens, parser = self.setup("parse_plotter")
        res = parser.parsePlotter()
        self.assertIsInstance(res.graph.graph, LineXYGraph)
        self.assertEqual(res.x.var.name, 'x')
        self.assertEqual(res.y.fun.fun, NumFunction.LOG)
        self.assertEqual(res.graph_name, 'line')

    def setup(self, input = None):
        def run_compile(content: str) -> Tokenizer:
            tk = Tokenizer(content)
            tk.tokenize()
            return tk
        if (input is not None):
            tk = read_program_file(f"tests/res/programs/{input}", run_compile)
        else:
            tk = read_program_file("tests/res/programs/example1", run_compile)
        print("GOT TOKENS")
        tokens = tk.tokens
        print(tokens)
        parser = Parser(tk)
        return tokens, parser


from .Node import Node
from mystix.language.shared.primitives import ConcreteValue
from mystix.language.shared.ast.visitor import Visitor


class Value(Node):

    def __init__(self, v: ConcreteValue):
        self.value = v

    def accept(self, v: Visitor):
        return v.visit_value(self)



from .Func import Func
from ..Var import Var
from .Operand import Operand
from ..Value import Value
from mystix.language.shared.ast.visitor import Visitor


class SimpleFunc(Func):

    def __init__(self, impacted_var: Var, op: Operand, v: Value):
        self.impacted_var: Var = impacted_var
        self.op: Operand = op
        self.rhs = v

    def accept(self, v: Visitor):
        return v.visit_simple_func(self)



from unittest import TestCase

from mystix.language.shared.primitives.graphs import ScatterXYGraph
from mystix.language.shared.primitives.numerical import NumFunction
from mystix.language.shared.primitives import Types
from tests.util.example_ast import simple_plot_example
import mystix.language.shared.ast as ast
from mystix.language.evaluation import Evaluator
from mystix.language.evaluation.errors import LanguageError
from mystix.language.evaluation.vars import UndefinedVariableError
from mystix.language.shared.primitives import values,  graphs


class GraphicsEvaluationTests(TestCase):

    """
    tests portions of evaluator with graphical side-effects
    This test should be ignored if graphics do not work on
    your system.
    """

    def test_plot_undefined_axis_no_source(self):
        """
        Should return without error
        """
        print("Testing undefined variable in Axis - no source")
        program = simple_plot_example()
        e = Evaluator(graphics=False)
        code, err = e.evaluate(program, duration=4000)
        self.assertEqual(0, code)

    def test_plot_undefined_axis(self):
        """
        Should briefly create a plot then error
        """
        print("Testing undefined variable in Axis")
        program = ast.Program(ast.Body([
            ast.Loader(ast.Var("source"),
                       ast.Source("https://covid-api.com/api/reports")),
            ast.Mapper(ast.Var("source"), "confirmed",
                       ast.Declare(ast.Type(Types.NUMBER), ast.Var("confirmed"))),
            ast.Plotter(ast.Graph(ScatterXYGraph()),
                        ast.VarAxis(ast.Var("t")),
                        ast.FuncAxis(ast.BuiltinFunc(NumFunction.SIN, ast.Var(
                            "t"))),
                        "sine_wave"),

        ]))
        e = Evaluator(graphics=False)
        code, err = e.evaluate(program, duration=4000)
        self.assertNotEqual(0, code)
        self.assertTrue(isinstance(err, UndefinedVariableError))

    """
    FIXME: running two tests consecutively with graphics enabled 
    creates seg-faults and QT errors
    """

    def test_plot_manual_define(self):
        """
        Should successfully create a plot and update with
        a constant value
        """
        print("Testing hard-coded variable value")
        program = simple_plot_example()
        e = Evaluator(graphics=False)
        e.env.extend('t', values.FloatValue(0.5))
        code, err = e.evaluate(program, duration=7000)
        self.assertEqual(0, code)

    def test_regular_program(self):
        """
        Should successfully create a plot and update with
        new values
        """
        print("Testing a full program")
        p = ast.Program(
            ast.Body([

                # source = live remote "https://covid-api.com/api/reports"
                ast.Loader(ast.Var("source"),
                           ast.Source("https://covid-api.com/api/reports")),

                # map source "confirmed" to number confirmed
                ast.Mapper(ast.Var("source"), "confirmed",
                           ast.Declare(ast.Type(Types.NUMBER), ast.Var("confirmed"))),

                # number count = 0
                ast.Assigner(ast.Declare(ast.Type(Types.NUMBER), ast.Var("count")),
                             ast.Value(values.IntegerValue(0))),

                # observe(source) do count++
                ast.Trigger(ast.Var("source"), ast.MathFuncs([ast.Increment(ast.Var(
                    "count"))])),

                # plot line_xy(count,confirmed) called "confirmed_cases"
                ast.Plotter(ast.Graph(graphs.LineXYGraph()),
                            ast.VarAxis(ast.Var("count")),
                            ast.VarAxis(ast.Var("confirmed")), "confirmed_cases"),

            ]))
        e = Evaluator(graphics=True)
        code, err = e.evaluate(p, duration=7000)
        self.assertEqual(0, code)



from .Command import Command
from ..Var import Var
from ..math_ast.MathFuncs import MathFuncs
from mystix.language.shared.ast.visitor import Visitor


class Trigger(Command):

    def __init__(self, var1: Var, math_funcs: MathFuncs):
        self.var1: Var = var1
        self.math_funcs: MathFuncs = math_funcs

    def accept(self, v: Visitor):
        return v.visit_trigger(self)



from PyQt5 import QtGui  # type: ignore
import pyqtgraph as pg  # type: ignore
from typing import cast, Callable


pg.setConfigOptions(antialias=True)


class GraphicsError(Exception):
    pass


class Graphics:

    def __init__(self):
        self.app = QtGui.QApplication([])
        self.windows = {}
        self.close_timer = pg.QtCore.QTimer()
        self.update_timer = pg.QtCore.QTimer()
        self.close_timer.timeout.connect(lambda: self.close())
        self.to_close: bool = False

        def check_close(instance):
            if instance.to_close:
                instance.app.closeAllWindows()
        self.add_update(lambda: check_close(self))

    def add_window(self, name: str, width: int = 500, height: int = 500) -> None:
        if name in self.windows:
            raise GraphicsError("Window '%s' already exists" % name)
        else:
            w = pg.GraphicsLayoutWidget(show=True, title=name)
            w.resize(width, height)
            self.windows[name] = w

    def add_update(self, p: Callable):
        self.update_timer.timeout.connect(p)

    def display(self, update_interval: int = 50, ttl=None) -> None:
        """
        :param update_interval:
        :param ttl: time before closing windows, in millis
        """
        if ttl is not None:
            self.close_timer.start(ttl)
        self.update_timer.start(update_interval)
        self.app.exec()

    def get_window(self, window_name: str) -> pg.GraphicsLayoutWidget:
        if window_name not in self.windows:
            raise GraphicsError("Window '%s' does not exist" % window_name)
        else:
            return cast(pg.GraphicsLayoutWidget, self.windows[window_name])

    def close(self):
        self.to_close = True



from typing import List

from ..Node import Node
from .Func import Func
from mystix.language.shared.ast.visitor import Visitor


class MathFuncs(Node):

    def __init__(self, function_list: List[Func]):
        self.mth_func_lst: List[Func] = function_list

    def accept(self, v: Visitor):
        return v.visit_math_funcs(self)



import unittest

from mystix.language.shared import ast, primitives


class ASTTests(unittest.TestCase):

    def test_program_construction(self):
        try:
            v: ast.Var = ast.Var("placeholder")
            t: ast.Type = ast.Type(primitives.Types.NUMBER)
            d: ast.Declare = ast.Declare(t,v)
            b: ast.Body = ast.Body([ast.commands_ast.Assigner(d,4), ast.commands_ast.Assigner(d,"hello")])
            p: ast.Program = ast.Program(b)
        except TypeError:
            self.fail("Failed to build")





class Graph:
    def equals(self, g) -> bool:
        return False


class LineXYGraph(Graph):
    def equals(self, g: Graph) -> bool:
        return isinstance(g, LineXYGraph)

    def __repr__(self):
        return "line"


class ScatterXYGraph(Graph):
    def equals(self, g: Graph) -> bool:
        return isinstance(g, ScatterXYGraph)

    def __repr__(self):
        return "scatter"



from enum import Enum


class NumFunction(Enum):
    LOG = "log"
    SIN = "sin"
    COS = "cos"
    EXP = "exp"


class NumOp(Enum):
    PLUS = "+"
    MINUS = "-"
    TIMES = "*"
    DIV = "/"
    EXP = "^"



from .Command import Command
from ..Var import Var
from ..Source import Source
from mystix.language.shared.ast.visitor import Visitor


class Loader(Command):

    def __init__(self, name: Var, source: Source):
        self.name: Var = name
        self.source: Source = source

    def accept(self, v: Visitor):
        return v.visit_loader(self)



from unittest import TestCase

from tests.util import ast_equal
from tests.util.example_ast import example_1
from mystix.language.shared import ast
from mystix.language.shared.primitives import Types
from mystix.language.shared.primitives.values import IntegerValue
from mystix.language.shared.primitives.graphs import ScatterXYGraph, LineXYGraph
from mystix.language.shared.primitives.numerical import NumFunction


class EquivalenceUtilTests(TestCase):

    def test_trivial(self):
        self.assertTrue(ast_equal(
            ast.Program(ast.Body([])), ast.Program(ast.Body([]))))

    def test_identical_construction(self):
        self.assertTrue(ast_equal(example_1(), example_1()))

    def test_identical_commands(self):
        self.assertTrue(
            ast_equal(
                ast.Program(
                    ast.Body([
                        ast.commands_ast.Loader(ast.Var("source"),
                                                ast.Source("www.coviddata.com/stream")),
                    ])),
        ast.Program(
            ast.Body([
                ast.commands_ast.Loader(ast.Var("source"),
                                        ast.Source("www.coviddata.com/stream")),
            ]))))

    def test_diff_num_commands(self):
        self.assertFalse(
            ast_equal(
                ast.Program(
                    ast.Body([
                        ast.commands_ast.Loader(ast.Var("source"),
                                                ast.Source("www.coviddata.com/stream")),
                    ])),
                ast.Program(
                    ast.Body([]))))

    def test_diff_commands(self):
        self.assertFalse(
            ast_equal(
                ast.Program(
                    ast.Body([
                        ast.commands_ast.Loader(ast.Var("source"),
                                                ast.Source("www.coviddata.com/stream")),
                    ])),
                ast.Program(
                    ast.Body([ast.Mapper(ast.Var("source"), "case_date",
                                         ast.Declare(ast.Type(Types.NUMBER), ast.Var("date")))]))))

    def test_multiple_commands(self):
        self.assertTrue(
            ast_equal(
                ast.Program(
                    ast.Body([
                        ast.commands_ast.Loader(ast.Var("source"),
                                                ast.Source("www.coviddata.com/stream")),
                        ast.Mapper(ast.Var("source"), "case_date",
                                   ast.Declare(ast.Type(Types.NUMBER), ast.Var("date")))
                    ])),
                ast.Program(
                    ast.Body([
                        ast.commands_ast.Loader(ast.Var("source"),
                                                ast.Source("www.coviddata.com/stream")),
                        ast.Mapper(ast.Var("source"), "case_date",
                                   ast.Declare(ast.Type(Types.NUMBER), ast.Var("date")))
                    ]))))

    def test_diff_order(self):
        self.assertFalse(
            ast_equal(
                ast.Program(
                    ast.Body([
                        ast.commands_ast.Loader(ast.Var("source"),
                                                ast.Source("www.coviddata.com/stream")),
                        ast.Mapper(ast.Var("source"), "case_date",
                                   ast.Declare(ast.Type(Types.NUMBER), ast.Var("date")))
                    ])),
                ast.Program(
                    ast.Body([
                        ast.Mapper(ast.Var("source"), "case_date",
                                   ast.Declare(ast.Type(Types.NUMBER), ast.Var(
                                       "date"))),
                        ast.commands_ast.Loader(ast.Var("source"),
                                                ast.Source("www.coviddata.com/stream")),
                    ]))))

    def test_different_load_var_names(self):
        self.assertFalse(
            ast_equal(
                ast.Program(
                    ast.Body([
                        ast.commands_ast.Loader(ast.Var("source"),
                                                ast.Source("www.coviddata.com/stream")),
                    ])),
                ast.Program(
                    ast.Body([
                        ast.commands_ast.Loader(ast.Var("source1"),
                                                ast.Source("www.coviddata.com/stream")),
                    ]))))

    def test_diff_load_url(self):
        self.assertFalse(
            ast_equal(
                ast.Program(
                    ast.Body([
                        ast.commands_ast.Loader(ast.Var("source"),
                                                ast.Source("https://www.coviddata.com/stream")),
                    ])),
                ast.Program(
                    ast.Body([
                        ast.commands_ast.Loader(ast.Var("source"),
                                                ast.Source("www.coviddata.com/stream")),
                    ]))))

    def test_diff_map_var(self):
        self.assertFalse(
            ast_equal(
                ast.Program(
                    ast.Body([
                        ast.Mapper(ast.Var("SOURCE"), "case_date",
                                   ast.Declare(ast.Type(Types.NUMBER), ast.Var("date")))
                    ])),
                ast.Program(
                    ast.Body([
                        ast.Mapper(ast.Var("source"), "case_date",
                                   ast.Declare(ast.Type(Types.NUMBER), ast.Var("date")))
                    ]))))

    def test_diff_map_field(self):
        self.assertFalse(
            ast_equal(
                ast.Program(
                    ast.Body([
                        ast.Mapper(ast.Var("source"), "case_date",
                                   ast.Declare(ast.Type(Types.NUMBER), ast.Var("date")))
                    ])),
                ast.Program(
                    ast.Body([
                        ast.Mapper(ast.Var("source"), "case_data",
                                   ast.Declare(ast.Type(Types.NUMBER), ast.Var("date")))
                    ]))))

    def test_diff_map_declare_type(self):
        self.assertFalse(
            ast_equal(
                ast.Program(
                    ast.Body([
                        ast.Mapper(ast.Var("source"), "case_date",
                                   ast.Declare(ast.Type(Types.NUMBER), ast.Var("date")))
                    ])),
                ast.Program(
                    ast.Body([
                        ast.Mapper(ast.Var("source"), "case_date",
                                   ast.Declare(ast.Type(Types.BINARY), ast.Var(
                                       "date")))
                    ]))))

    def test_diff_map_declare_var(self):
        self.assertFalse(
            ast_equal(
                ast.Program(
                    ast.Body([
                        ast.Mapper(ast.Var("source"), "case_date",
                                   ast.Declare(ast.Type(Types.NUMBER), ast.Var("date")))
                    ])),
                ast.Program(
                    ast.Body([
                        ast.Mapper(ast.Var("source"), "case_date",
                                   ast.Declare(ast.Type(Types.NUMBER), ast.Var(
                                       "count")))
                    ]))))

    def test_diff_math_funcs_(self):
        self.assertFalse(
            ast_equal(
                ast.Program(
                    ast.Body([
                        ast.Trigger(ast.Var("source"), ast.MathFuncs([ast.Increment(ast.Var(
                            "count"))]))
                    ])),
                ast.Program(
                    ast.Body([
                        ast.Trigger(ast.Var("source"), ast.MathFuncs([ast.Decrement(ast.Var(
                            "count"))]))
                    ]))))

    def test_diff_num_math_funcs_(self):
        self.assertFalse(
            ast_equal(
                ast.Program(
                    ast.Body([
                        ast.Trigger(ast.Var("source"), ast.MathFuncs([ast.Increment(ast.Var(
                            "count"))]))
                    ])),
                ast.Program(
                    ast.Body([
                        ast.Trigger(ast.Var("source"), ast.MathFuncs([ast.Increment(ast.Var(
                            "count")), ast.Decrement(ast.Var(
                            "count"))]))
                    ]))))



from .Node import Node
from mystix.language.shared.primitives import Types
from mystix.language.shared.ast.visitor import Visitor


class Type(Node):

    def __init__(self, t: Types):
        self.type = t

    def accept(self, v: Visitor):
        return v.visit_type(self)



from .Func import Func
from ..Var import Var
from mystix.language.shared.ast.visitor import Visitor


class FastFunc(Func):

    def __init__(self, impacted_var: Var):
        self.impacted_var: Var = impacted_var


class Increment(FastFunc):

    def accept(self, v: Visitor):
        return v.visit_increment(self)


class Decrement(FastFunc):

    def accept(self, v: Visitor):
        return v.visit_decrement(self)



"""
To provide type checking for this file (and subclasses of Visitor)
and avoid a circular import between Node types and the Visitor class
itself, we do a dynamic import only in the case of static type
checking.
Per this post:
https://stackoverflow.com/questions/39740632/python-type-hinting-without-cyclic-imports

To implement the visitor pattern structure, we set this file as the "interface"
in an informal structure.
Using this post for reference:
https://refactoring.guru/design-patterns/visitor/python/example
"""
from __future__ import annotations
from typing import TYPE_CHECKING as STATIC_CHECK
if STATIC_CHECK:
    from mystix.language.shared.ast import *
from abc import ABC, abstractmethod


class Visitor(ABC):
    @abstractmethod
    def visit_program(self, p: Program):
        pass

    @abstractmethod
    def visit_body(self, b: Body):
        pass

    @abstractmethod
    def visit_loader(self, l: Loader):
        pass

    @abstractmethod
    def visit_mapper(self, m: Mapper):
        pass

    @abstractmethod
    def visit_declare(self, d: Declare):
        pass

    @abstractmethod
    def visit_assigner(self, a: Assigner):
        pass

    @abstractmethod
    def visit_trigger(self, t: Trigger):
        pass

    @abstractmethod
    def visit_plotter(self, pltr: Plotter):
        pass

    @abstractmethod
    def visit_var(self, v: Var):
        pass

    @abstractmethod
    def visit_source(self, s: Source):
        pass

    @abstractmethod
    def visit_type(self, t: Type):
        pass

    @abstractmethod
    def visit_value(self, v: Value):
        pass

    @abstractmethod
    def visit_graph(self, g: Graph):
        pass

    @abstractmethod
    def visit_var_axis(self, va: VarAxis):
        pass

    @abstractmethod
    def visit_func_axis(self, fa: FuncAxis):
        pass

    @abstractmethod
    def visit_math_funcs(self, mf: MathFuncs):
        pass

    @abstractmethod
    def visit_simple_func(self, smp: SimpleFunc):
        pass

    @abstractmethod
    def visit_increment(self, i: Increment):
        pass

    @abstractmethod
    def visit_decrement(self, d: Decrement):
        pass

    @abstractmethod
    def visit_builtin_func(self, bltn: BuiltinFunc):
        pass

    @abstractmethod
    def visit_operand(self, op: Operand):
        pass


from mystix.language.shared.primitives.values import (
    NumericalValue, FloatValue, IntegerValue )
from mystix.language.shared.primitives.numerical import NumOp, NumFunction
from mystix.util.errors import NonExhaustiveTypeCaseError
from typing import cast

import math


class MathError(Exception):
    pass


def safe_divide(a, b):
    if b == 0:
        raise MathError("Cannot divide by zero!")
    else:
        return a / b


operations = dict({
    NumOp.PLUS: lambda a, b: a + b,
    NumOp.MINUS: lambda a, b: a - b,
    NumOp.TIMES: lambda a, b: a * b,
    NumOp.DIV: lambda a, b: safe_divide(a,b),
    NumOp.EXP: lambda a, b: a**b
})

functions = dict({
    NumFunction.LOG: lambda x: math.log(x),
    NumFunction.SIN: lambda x: math.sin(x),
    NumFunction.COS: lambda x: math.cos(x),
    NumFunction.EXP: lambda x: math.exp(x)
})

quick_func = dict({
    "inc": lambda x: x + 1,
    "dec": lambda x: x - 1
})


def apply_op(o: NumOp, a: NumericalValue, b: NumericalValue) -> NumericalValue:
    if o in operations:
        result = operations[o](a.value, b.value)
        if type(a) is IntegerValue and type(b) is IntegerValue:
            return cast(NumericalValue, IntegerValue(int(result)))
        else:
            return cast(NumericalValue, FloatValue(float(result)))
    else:
        raise NonExhaustiveTypeCaseError()


def apply_fn(f: NumFunction, x: NumericalValue) -> FloatValue:
    if f in functions:
        try:
            return FloatValue(float(functions[f](x.value)))
        except ValueError:
            raise MathError(
                "Error executing function %s with input %s" % (f,str(x.value)))

    else:
        raise NonExhaustiveTypeCaseError()


def apply_qk(f: str, x: NumericalValue) -> NumericalValue:
    result = quick_func[f](x.value)
    if type(x) is IntegerValue:
        return cast(NumericalValue, IntegerValue(int(result)))
    else:
        return cast(NumericalValue, FloatValue(float(result)))


from .Command import Command
from ..Value import Value
from ..Declare import Declare
from mystix.language.shared.ast.visitor import Visitor


class Assigner(Command):

    def __init__(self, decl: Declare, value: Value):
        self.decl = decl
        self.value: Value = value

    def accept(self, v: Visitor):
        return v.visit_assigner(self)



from __future__ import annotations
from typing import TYPE_CHECKING as STATIC_CHECK
if STATIC_CHECK:
    from mystix.language.shared.ast import *
from mystix.language.shared.ast.visitor import Visitor


class ProgramPrinter(Visitor):

    def print(self, p: Program) -> str:
        return self.visit_program(p)

    def visit_program(self, p: Program) -> str:
        return "(Program %s)" % p.body.accept(self)

    def visit_body(self, b: Body) -> str:
        commands = [c.accept(self) for c in b.commands]
        return "(Body %s)" % " ".join(commands)

    def visit_loader(self, l: Loader) -> str:
        return "(Loader %s %s)" % \
            (l.name.accept(self), l.source.accept(self))

    def visit_mapper(self, m: Mapper) -> str:
        return "(Mapper %s '%s' %s)" % \
            (m.src.accept(self), m.tbl_field, m.decl.accept(self))

    def visit_declare(self, d: Declare) -> str:
        return "(Declare %s %s)" % \
            (d.type.accept(self), d.var.accept(self))

    def visit_assigner(self, a: Assigner) -> str:
        return "(Assigner %s %s)" % \
            (a.decl.accept(self), a.value.accept(self))

    def visit_trigger(self, t: Trigger) -> str:
        return "(Trigger %s %s)" % \
            (t.var1.accept(self), t.math_funcs.accept(self))

    def visit_plotter(self, pltr: Plotter) -> str:
        return "(Plotter %s %s %s '%s')" % \
            (pltr.graph.accept(self),
             pltr.x.accept(self),
             pltr.y.accept(self), pltr.graph_name)

    def visit_var(self, v: Var) -> str:
        return "(Var '%s')" % v.name

    def visit_source(self, s: Source) -> str:
        return "(Source '%s')" % s.url

    def visit_type(self, t: Type) -> str:
        return "(Type %s)" % t.type

    def visit_value(self, v: Value) -> str:
        return "(Value %s)" % repr(v.value)

    def visit_graph(self, g: Graph) -> str:
        return "(Graph '%s')" % repr(g.graph)

    def visit_var_axis(self, va: VarAxis) -> str:
        return "(VarAxis %s)" % va.var.accept(self)

    def visit_func_axis(self, fa: FuncAxis) -> str:
        return "(FuncAxis %s)" % fa.fun.accept(self)

    def visit_math_funcs(self, mf: MathFuncs) -> str:
        return "(%s)" % \
            " ".join([f.accept(self) for f in mf.mth_func_lst])

    def visit_simple_func(self, smp: SimpleFunc) -> str:
        return "(%s %s %s)" % \
            (smp.op.accept(self),
             smp.impacted_var.accept(self), smp.rhs.accept(self))

    def visit_increment(self, i: Increment) -> str:
        return "(Increment %s)" % i.impacted_var.accept(self)

    def visit_decrement(self, d: Decrement) -> str:
        return "(Decrement %s)" % d.impacted_var.accept(self)

    def visit_builtin_func(self, bltn: BuiltinFunc) -> str:
        return "(BuiltinFunc %s %s)" % \
            (bltn.fun, bltn.input.accept(self))

    def visit_operand(self, op: Operand) -> str:
        return "%s" % op.op



from .Node import Node
from .Body import Body
from mystix.language.shared.ast.visitor import Visitor


class Program(Node):
    """
    Represents a program. This contains no additional fields other than
    the program body but is kept here for consistency with the Grammar
    """
    def __init__(self, b: Body):
        self.body: Body = b
        

    def accept(self, v: Visitor):
        return v.visit_program(self)



from .Node import Node
from mystix.language.shared.ast.visitor import Visitor


class Source(Node):

    def __init__(self, url: str):
        self.url: str = url

    def accept(self, v: Visitor):
        return v.visit_source(self)
